import contains from 'lodash.contains';
import BBPromise from 'bluebird';
import pg from 'pg';
BBPromise.promisifyAll(pg);
import logger from 'debug';
var debug = logger('postgresql.migrator');

const VERSION_SEPERATOR = '@';

var capitalizeFirstLetter = (string) => {
  return string.charAt(0).toUpperCase() + string.slice(1);
}

/**
 * PostgresqlMigrator
 *
 * @class PostgresqlMigrator
 */
class PostgresqlMigrator {
  /**
   * @param {PostgresqlAdapter} adapter
   */
  constructor(adapter, config) {
    this.adapter = adapter;
    this.config = config;
  }

  /**
   * Create a new postgresql database and it's tables.
   * @param  {String} projectId
   * @param  {Object} oldProject
   * @param  {Object} project
   * @param  {Array} changes
   *
   * @return {Promise}
   */
  async migrate(projectId, oldProject, project, changes) {
    var connectionString = 'postgresql://' + this.config.user + ':' + this.config.password + '@' + this.config.host + '/';
    await this.checkExtensions(connectionString);
    await this.createDatabase(connectionString, project.name, project.version);
    let client = await this.adapter.getConnection(project.name, project.version);
    await this.createTables(client, project.schemas);
    return {
      success: true
    };
  }

  checkExtensions(connectionString) {
    return pg.connectAsync(connectionString + 'template_postgis')
      .spread((postgisTemplateConnection, done) => {
        return postgisTemplateConnection.queryAsync('SELECT extname, extversion from pg_extension')
          .then((result) => {
            var promise = BBPromise.resolve();
            var extensions = result.rows.map((row) => row.extname);
            if ( ! contains(extensions, 'postgis')) {
              debug('"postgis" extension missing, installing it.');
              promise = promise.then(() => {
                return postgisTemplateConnection.queryAsync('CREATE EXTENSION postgis');
              });
            }
            if ( ! contains(extensions, 'uuid-ossp')) {
              debug('"uuid-ossp" extension missing, installing it.');
              promise = promise.then(() => {
                return postgisTemplateConnection.queryAsync('CREATE EXTENSION "uuid-ossp"');
              });
            }
            return promise;
          })
          .finally(() => {
            return done();
          });
      });
  }

  /**
   * Creates a new database for given projectId and project version.
   *
   * @param  {String} connectionString The postgresql connectionString
   * @param  {String} projectName      The project name
   * @param  {Number} projectVersion   The project version
   */
  createDatabase(connectionString, projectName, projectVersion) {
    return pg.connectAsync(connectionString + 'template_postgis')
      .spread((postgresConnection, done) => {
        return postgresConnection.queryAsync(`CREATE DATABASE "${projectName}${VERSION_SEPERATOR}${projectVersion}" TEMPLATE=template_postgis`)
          .catch((err) => {
            console.error('Error while creating database', err);
            throw err;
          })
          .finally(() => {
            return done();
          });
      });
  }

  /**
   * Creates tables in database based on the given blueprints.
   *
   * @param  {PostgresqlClient}  client
   * @param  {Object}            schemas
   */
  createTables(client, schemas) {
    return BBPromise.map(Object.keys(schemas), (schemaId) => {
      let schema = schemas[schemaId];
      return client.schema.createTable(schema.name, (table) => {
        Object.keys(schema.columns).forEach((columnId) => {
          let column = schema.columns[columnId];
          let chain = this['define' + capitalizeFirstLetter(column.type).replace('[]', 'Array')](table, column.name);
          if (schema.primary === columnId) {
            chain.primary();
          }
          if (column.unique) {
            chain.unique();
          }
          if (column.references) {
            chain.inTable(schemas[column.references.schemaId].name)
              .references(schemas[column.references.schemaId].columns[column.references.columnId].name);
          }
          if (column.unsigned) {
            chain.unsigned();
          }
          if (column.nullable === false) {
            chain.notNullable();
          }
          if (column.comment) {
            chain.comment(column.comment);
          }
        });
      })
      .then((res) => {
        return res;
      }, (err) => {
        console.error('error while creating tables', err);
      });
    });
  }

  /**
   * Defines an INTEGER column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineInteger(table, columnName) {
    return table.integer(columnName);
  }

  /**
   * Defines an INTEGER[] column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineIntegerArray(table, columnName) {
    return table.specificType(columnName, 'INTEGER[]');
  }

  /**
   * Defines an UUID column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineUuid(table, columnName) {
    return table.uuid(columnName);
  }

  /**
   * Defines an UUID[] column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineUuidArray(table, columnName) {
    return table.specificType(columnName, 'UUID[]');
  }

  /**
   * Defines an STRING column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineString(table, columnName) {
    return table.string(columnName);
  }

  /**
   * Defines an STRING[] column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineStringArray(table, columnName) {
    return table.specificType(columnName, 'VARCHAR[]');
  }

  /**
   * Defines an TEXT column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineText(table, columnName) {
    return table.text(columnName);
  }

  /**
   * Defines an TEXT[] column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineTextArray(table, columnName) {
    return table.specificType(columnName, 'TEXT[]');
  }

  /**
   * Defines an DATETIME column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineDatetime(table, columnName) {
    return table.dateTime(columnName);
  }

  /**
   * Defines an DATETIME[] column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineDatetimeArray(table, columnName) {
    return table.specificType(columnName, 'TIMESTAMP WITH TIME ZONE[]');
  }

  /**
   * Defines an DATE column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineDate(table, columnName) {
    return table.date(columnName);
  }

  /**
   * Defines an DATE[] column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineDateArray(table, columnName) {
    return table.specificType(columnName, 'DATE[]');
  }

  /**
   * Defines an REAL column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineFloat(table, columnName) {
    return table.float(columnName);
  }

  /**
   * Defines an REAL[] column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineFloatArray(table, columnName) {
    return table.specificType(columnName, 'REAL[]');
  }

  /**
   * Defines an GEOMETRY column of type Point on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  definePoint(table, columnName) {
    return table.specificType(columnName, 'geometry(Point,4326)');
  }

  /**
   * Defines an GEOMETRY column of type MultiPoint on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  definePointArray(table, columnName) {
    return table.specificType(columnName, 'geometry(MultiPoint,4326)');
  }

  /**
   * Defines an GEOMETRY column of type LineString on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineLinestring(table, columnName) {
    return table.specificType(columnName, 'geometry(LineString,4326)');
  }

  /**
   * Defines an GEOMETRY column of type MultiLineString on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineLinestringArray(table, columnName) {
    return table.specificType(columnName, 'geometry(MultiLineString,4326)');
  }

  /**
   * Defines an GEOMETRY column of type Polygon on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  definePolygon(table, columnName) {
    return table.specificType(columnName, 'geometry(Polygon,4326)');
  }

  /**
   * Defines an GEOMETRY column of type MultiPolygon on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  definePolygonArray(table, columnName) {
    return table.specificType(columnName, 'geometry(MultiPolygon,4326)');
  }

  /**
   * Defines an BOOLEAN column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineBoolean(table, columnName) {
    return table.boolean(columnName);
  }

  /**
   * Defines an BOOLEAN[] column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineBooleanArray(table, columnName) {
    return table.specificType(columnName, 'BOOLEAN[]');
  }

  /**
   * Defines a JSON column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineJson(table, columnName) {
    return table.json(columnName);
  }

  /**
   * Defines a JSON[] column on given table.
   *
   * @protected
   * @param  {Object} table     Knex.js table object (given through createTable's callback)
   * @param  {String} columnName Name of the column
   */
  defineJsonArray(table, columnName) {
    return table.specificType(columnName, 'JSON[]');
  }

}

export default PostgresqlMigrator;
