import contains  from 'lodash.contains';
import BBPromise from 'bluebird';
var knex = require('knex')({dialect: 'postgres'});

/**
 * PostgresqlRepository
 *
 * @class PostgresqlReposirory
 */
export default class PostgresqlRepository {
  /**
   * @param {PostgresqlSerializer}  serializer
   * @param {PostgresqlAdapter}     adapter
   */
  constructor(serializer, adapter) {
    this.serializer = serializer;
    this.adapter = adapter;
  }

  /**
   * Insert / update item.
   *
   * @param  {Item} item          The item to store
   * @param  {Schema} schema      The schema for the name and primary properties.
   * @param  {Metadata} metadata  An object with a projectName and projectVersion property.
   * @return {Promise}
   */
  async putItem(item, schema, metadata) {
    // serialize the item
    let serializedItem = this.serializer.serialize(item, schema);
    // retrieve the knex.js client for the given version
    let client = this.adapter.getConnection(metadata.projectName, metadata.projectVersion);
    // check if the item exists
    let exists = await this.itemExists(client, schema, serializedItem);
    if (exists) {
      // update the item
      return await this.updateItem(client, schema, serializedItem);
    } else {
      // insert the item
      return await this.insertItem(client, schema, serializedItem);
    }
  }

  /**
   * Remove an item.
   * @param  {String}   itemId      The id of the item
   * @param  {Schema}   schema      The schema for the name and primary properties.
   * @param  {Metadata} metadata    An object with a projectName and projectVersion property.
   * @return {Promise}
   */
  delItem(itemId, schema, metadata) {
    let client = this.adapter.getConnection(metadata.projectName, metadata.projectVersion);
    var wheres = {};
    var key = schema.columns[schema.primary].name;
    wheres[key] = itemId;
    return client.table(schema.name)
      .where(wheres)
      .del()
      // triggers knex.js's promise strategy
      .then((res) => res);
  }

  /**
   * Checks if an item already exists in the database.
   * @protected
   * @param  {PostgresqlClient} client          The knex client
   * @param  {Schema}           schema          Schema options
   * @param  {Item}             item            The serialized item
   * @return {Promise} Promise resolving true or false (exists or doesn't exist).
   */
  itemExists(client, schema, item) {
    return client.table(schema.name)
      .first()
      .where(this.getWheres(schema, item))
      .then((row) => !!row);
  }

  /**
   * Updates an item in the database.
   * @protected
   * @param  {PostgresqlClient} client            The knex client
   * @param  {Schema}           schema            Schema information
   * @param  {Item}             item              The serialized item
   * @return {Promise}
   */
  updateItem(client, schema, item) {
    return client.table(schema.name)
      .where(this.getWheres(schema, item))
      .update(item)
      // triggers knex.js's promise strategy
      .then((res) => res);
  }

  /**
   * Inserts a new item into the database.
   * @protected
   * @param  {PostgresqlClient} client          The knex client
   * @param  {Schema}           schema          Schema information
   * @param  {Item}             item            The serialized item
   * @return {Promise}
   */
  insertItem(client, schema, item) {
    return client.table(schema.name)
      .insert(item)
      // triggers knex.js's promise strategy
      .then((res) => res);
  }

  /**
   * Get where query
   * @protected
   * @param  {Schema} schema         Schema information
   * @param  {Item}   item           The serialized item
   * @return {Object}                The query
   */
  getWheres(schema, item) {
    var wheres = {};
    var key = schema.columns[schema.primary].name;
    wheres[key] = item[key];
    return wheres;
  }

}
