import pg from 'pg';
import contains from 'lodash.contains';
import through from 'through';
import BBPromise from 'bluebird';
import transformer from 'schema-mapper-transformer';
import QueryStream from 'pg-query-stream';
import logger from 'debug';
BBPromise.promisifyAll(pg);
var debug = logger('postgresql.syncer');

const VERSION_SEPERATOR = '@';

class PostgresqlSyncer {

  constructor(config, adapter, serializer) {
    this.config = config;
    this.adapter = adapter;
    this.serializer = serializer;
  }

  /**
   * Sync data from the old database to the new one,
   * transforming items to the new version along the way.
   *
   * @param  {String} projectId
   * @param  {Project} oldProject
   * @param  {Project} project
   * @param  {Changes} changes
   * @return {Promise}
   */
  async sync(projectId, oldProject, project, changes) {
    debug('-> sync');
    var self = this;
    let fromConnectionString = 'postgresql://' + this.config.user + ':' + this.config.password + '@' + this.config.host + '/' + oldProject.name + VERSION_SEPERATOR + oldProject.version;
    var fromClient = await this.getConnection(fromConnectionString);
    fromClient.connection.stream.setMaxListeners(100);
    // retrieve the knex.js client for the new project version
    var toClient = this.adapter.getConnection(project.name, project.version);
    return BBPromise.each(Object.keys(oldProject.schemas), (schemaId) => {
      var oldSchema = oldProject.schemas[schemaId];
      var schema = project.schemas[schemaId];
      if (!schema) {
        return;
      }
      let transformItem = through(function(item) {
        Object.keys(oldSchema.columns).forEach((columnId) => {
          let oldColumn = oldSchema.columns[columnId];
          if (self.isGeoField(oldColumn.type)) {
            item[oldColumn.name] = JSON.parse(item[oldColumn.name]);
          }
        });
        let transformedItem = transformer.transformItem(item, changes);
        if (transformedItem !== null) {
          this.emit('data', transformedItem);
        }
      });

      var toStream = through(function(transformedItem) {
        debug('-> sync.insert');
        var streamSelf = this;
        let serializedItem = self.serializer.serialize(transformedItem, schema);
        streamSelf.pause();
        toClient.table(schema.name)
          .insert(serializedItem)
          // triggers knex.js's promise strategy
          .then((res) => {
            streamSelf.resume();
            return res;
          })
          .catch((err) => {
            throw err;
          });
      });

      return new BBPromise((resolve, reject) => {
        var query = new QueryStream('SELECT ' + this.schemaToSelects(oldSchema) + ' FROM "' + oldSchema.name + '"');
        var fromStream = fromClient.query(query);
        fromStream
          .pipe(transformItem)
          .pipe(toStream)
          .on('end', () => {
            debug('-> sync.done');
            resolve();
          })
          .on('error', (err) => {
            debug('-> sync.doneWithErrors', err);
            reject(err);
          });
      });
    });
  }

  getConnection(connectionString) {
    return pg.connectAsync(connectionString)
      .spread((client, done) => {
        this.connectionCloser = done;
        return client;
      });
  }

  schemaToSelects(schema) {
    var selects = ['*'];

    Object.keys(schema.columns).forEach((columnId) => {
      let column = schema.columns[columnId];

      if (this.isGeoField(column.type)) {
        selects.push('ST_AsGeoJson(' + column.name + ') as ' + column.name);
      }

      if (column.type === 'datetime') {
        selects.push('to_char(' + column.name + ', \'yyyy-MM-dd hh:mi:ss\') as ' + column.name);
      }

      if (column.type === 'date') {
        selects.push('to_char(' + column.name + ', \'yyyy-MM-dd\') as ' + column.name);
      }
    });

    return selects.join(',');
  }

  isGeoField(type) {
    return contains(['linestring', 'linestring[]', 'point', 'point[]', 'polygon', 'polygon[]'], type);
  }

}

export default PostgresqlSyncer;
